<?php


/**
 * Client class.
 *
 * TODO: also implements CRMRemoteQueryInterface or somesuch!
 */
class clients_connection_ms_dynamics_soap extends clients_connection_base
  implements
    ClientsConnectionAdminUIInterface,
    ClientsRemoteEntityInterface {

  // ================================= Clients admin UI

  /**
   * Declare an array of properties which should be treated as credentials.
   *
   * This lets the credentials storage plugin know which configuration
   * properties to take care of.
   *
   * @return
   *  A flat array of property names.
   */
  function credentialsProperties() {
    return array('wlid_username', 'wlid_password');
  }

  /**
   * Extra form elements specific to a class's edit form.
   *
   * This is the same pattern as node_form() -- just ignore the object behind
   * the curtain ;)
   *
   * @param $form_state
   *  The form state from the main form, which you probably don't need anyway.
   *
   * @see clients_connection_form()
   * @see clients_connection_form_submit()
   */
  function connectionSettingsFormAlter(&$form, &$form_state) {
    // Change the endpoint element.
    $form['endpoint']['#title'] = t('MS CRM Service endpoint');
    $form['endpoint']['#description'] = t('MS CRM Service endpoint URL');

    $form['configuration']['application_name'] = array(
      '#type' => 'textfield',
      '#title' => t('Name of the application'),
      '#size' => 50,
      '#maxlength' => 50,
      '#description' => t('The name of the application connecting to MS CRM'),
      // TODO: uncomment these when out of dev!
      //'#required' => TRUE,
    );
    $form['credentials']['wlid_username'] = array(
      '#type' => 'textfield',
      '#title' => t('Windows Live username'),
      '#size' => 50,
      '#maxlength' => 100,
      '#description' => t('The windows live account username for connecting to MS CRM'),
      //'#required' => TRUE,
    );
    $form['credentials']['wlid_password'] = array(
      '#type' => 'textfield',
      '#title' => t('Windows Live password'),
      '#size' => 50,
      '#maxlength' => 100,
      '#description' => t('The windows live account password for connecting to MS CRM'),
      //'#required' => TRUE,
    );
    $form['configuration']['device_username'] = array(
      '#type' => 'textfield',
      '#title' => t('Device username'),
      '#size' => 30,
      '#maxlength' => 30,
      '#description' => t('The username of the device connecting to MS CRM'),
      //'#required' => TRUE,
    );
    $form['configuration']['device_password'] = array(
      '#type' => 'textfield',
      '#title' => t('Device password'),
      '#size' => 30,
      '#maxlength' => 30,
      '#description' => t('The password of the device connecting to MS CRM'),
      //'#required' => TRUE,
    );
    $form['configuration']['read_only'] = array(
      '#type' => 'checkbox',
      '#title' => t('Read-only'),
      '#description' => t('Marks this connection read-only. This requires associated Resources to support this feature.'),
    );
  }

  // ================================= ClientsRemoteEntityInterface

  /**
   * Load a remote entity.
   *
   * The resource calling this should take care of process the data returned by
   * this into a Drupal entity.
   *
   * @param $entity_type
   *  The entity type to load.
   * @param $id
   *  The (remote) ID of the entity.
   *
   * @return
   *  The remote entity data.
   */
  function remote_entity_load($entity_type, $id) {
    $query = $this->getRemoteEntityQuery('select');
    $query->base($entity_type);
    $query->entityCondition('entity_id', $id);
    $result = $query->execute();
    //dsm($result, 'remote_entity_load $result');

    // There's only one. Same pattern as entity_load_single().
    return reset($result);
  }

  /**
   * Load multiple remote entities.
   *
   * The resource calling this should take care of process the data returned by
   * this into Drupal entities.
   *
   * @param $entity_type
   *  The entity type to load.
   * @param $id
   *  An array of the (remote) entity IDs.
   *
   * @return
   *  The remote entity data. This is a numeric array in the same order as the
   *  given array of remote IDs (at least, observation shows that this is the
   *  case so far! Who knows with Microsoft!).
   */
  function remote_entity_load_multiple($entity_type, $ids) {
    $query = $this->getRemoteEntityQuery('select');
    $query->base($entity_type);
    $query->entityCondition('entity_id', $ids, 'In');
    $result = $query->execute();

    return $result;
  }

  /**
   * Save a remote entity.
   *
   * @param $entity_type
   *  The entity type to save.
   * @param $entity
   *  The entity to save.
   *
   * @return
   *  If the entity is being created remotely, the new remote GUID.
   */
  function remote_entity_save($entity_type, $entity, $remote_properties = array()) {
    // Determine whether this is an insert or an update. An entity not yet
    // saved remotely will have no remote ID property.
    $action = empty($entity->remote_id) ? 'insert' : 'update';
    //dsm($action, '$action');

    $query = $this->getRemoteEntityQuery($action);
    $query->setEntity($entity_type, $entity);

    if ($action == 'insert') {
      $guid = $query->execute();
      return $guid;
    }
    else {
      $query->addFields($remote_properties);

      $query->execute();
    }
  }

  /**
   * Save multiple entities remotely, using XRM's bulk update.
   *
   * @param $entity_type
   *  The entity type to save.
   * @param $entities
   *  An array of entities to save in bulk, keyed by the entity ID.
   * @param $remote_properties
   *  (Optional) An array of properties to save. Values should be names of
   *  properties which are keys in the entity info 'property map' array. Only
   *  applies when updating rather than inserting.
   *
   * @return
   *  An array of the remote GUIDs created by the operation, keyed by the
   *  entity ID. Thus an entity which is updated rather than inserted will not
   *  be present in the array.
   */
  function remote_entity_save_multiple($entity_type, $entities, $remote_properties = array()) {
    $query = $this->getRemoteEntityQuery('bulk_save');
    $query->setEntities($entity_type, $entities);
    $query->addFields($remote_properties);

    // Execute the query. The guids array is keyed by entity ID.
    $guids = $query->execute();

    return $guids;
  }

  /**
   * Provide a map of remote property types to Drupal types.
   *
   * Roughly analogous to _entity_metadata_convert_schema_type().
   *
   * @return
   *  An array whose keys are remote property types as used as types for fields
   *  in hook_remote_entity_query_table_info(), and whose values are types
   *  recognized by the Entity Metadata API (as listed in the documentation for
   *  hook_entity_property_info()).
   *  If a remote property type is not listed here, it will be mapped to 'text'
   *  by default.
   */
  function entity_property_type_map() {
    return array(
      'EntityCollection' => 'list<string>',
    );
  }

  /**
   * Provide setter and getter callbacks for specific remote property types.
   *
   * @return
   *  An array keyed by remote property type, where each value is an array
   *  defining the callbacks for remote properties of this type. Keys in this
   *  must be 'getter callback' and 'setter callback'. The raw setter and getter
   *  callbacks may also be included.
   *
   * @see RemoteEntityAPIDefaultMetadataController::entityPropertyInfo()
   */
  function entity_property_type_callbacks() {
    return array(
      'dateTime' => array(
        'getter callback' => 'clients_ms_dynamics_soap_entity_mapped_property_datetime_get',
        'setter callback' => 'clients_ms_dynamics_soap_entity_mapped_property_datetime_set',
        // Add callbacks for raw values, which can just be the basic ones.
        'raw getter callback' => 'remote_entity_entity_mapped_property_get',
        'raw setter callback' => 'remote_entity_entity_mapped_property_set',
      ),
    );
  }

  /**
   * Get a new RemoteEntityQuery object appropriate for the connection.
   *
   * @param $query_type
   *  (optional) The type of the query. Defaults to 'select'.
   *
   * @return
   *  A remote query object of the type appropriate to the query type.
   */
  function getRemoteEntityQuery($query_type = 'select') {
    // TODO: change class names to be determined automatically, like in core?
    switch ($query_type) {
      case 'select':
        return new MSDynamicsSoapSelectQuery($this);
      case 'fetch_expression':
        return new MSDynamicsSoapSelectFetchExpressionQuery($this);
      case 'insert':
        return new MSDynamicsSoapInsertQuery($this);
      case 'update':
        return new MSDynamicsSoapUpdateQuery($this);
      case 'bulk_save':
        return new MSDynamicsSoapBulkSaveQuery($this);
      case 'set_state':
        return new MSDynamicsSoapSetStateRequestQuery($this);
    }
  }

  /**
   * Helper to create a dummy remote ID, to use in testing.
   *
   * XRM uses GUIDS of the form:
   *  00000000-0000-0000-0000-000000000000
   * We create a dummy of the form:
   *  00000000-0000-FFFF-FFFF-iiiitttttttt
   * where:
   *  - 0 and F are literals. These make our dummy GUIDs in theory easily
   *    recognizable as such, to help prevent head-scratching (such as when
   *    dummy-saved entities fail to update from remote).
   *  - i is an incremented number to ensure uniqueness within a single request
   *    (if you need to create more than 9999 remote entities in a single
   *    request, contact me with your server spec).
   *  - t is the request timestamp converted to hex.
   */
  function generateDummyRemoteID() {
    static $i;
    $i++;

    $prefix = '00000000-0000-FFFF-FFFF';

    $increment = sprintf("%04d", $i);

    // Current timestamps converted to hexadecimal yield 8 characters, and will
    // continue to do so until Sun, 07 Feb 2106 06:28:15 (and the 2038 problem
    // will happen way before then).
    $date_hex = dechex(time());

    $dummy_guid = "$prefix-$increment$date_hex";

    return $dummy_guid;
  }

  // ================================= ClientsCRMQueryInterface

  // More general querying methods?
  // TODO!

  // ================================= SOAP API

  /**
   * Cache lifetime: 1 hour.
   *
   * XRM caches our credentials for 1 day.
   */
  const CACHE_LIFETIME = 3600;

  /**
   * Connect and log in to the remote service.
   */
  function connect() {
    // Load the library now, as we need the CrmAuthenticationHeader class to
    // exist in order for the cache data to be properly unserialized.
    libraries_load('CRMSoapAuthPHP');

    $cid = 'clients_connection_ms_dynamics_soap:' . $this->name;
    $cache = cache_get($cid);
    if ($cache === FALSE || REQUEST_TIME > $cache->expire) {
      // login() will set the cache.
      $this->login();
    }
    else {
      $this->authHeader = $cache->data;
      $this->isLoggedIn = TRUE;
    }
  }

  /**
   * Try to authenticate with MS Dynamics.
   *
   * Helper for connect().
   */
  function login() {
    $this->credentialsLoad();

    // Connect using the library.
    libraries_load('CRMSoapAuthPHP');
    $crmAuth = new CrmAuth();
    $authHeader = $crmAuth->GetHeaderOnline($this->credentials['wlid_username'], $this->credentials['wlid_password'], $this->endpoint);

    // We expect the Expires property to be set.
    if (empty($authHeader->Expires)) {
      // TODO: the CRMSoapAuthPHP library doesn't handle a bad password gracefully
      // and outputs warnings!
      throw new Exception("Invalid login credentials.");
    }

    // Set the authentication header for subsequent requests.
    $this->authHeader = $authHeader;

    // Expiry date is of form: 2016-03-18T13:54:31Z
    $expiry_date = DateTime::createFromFormat(DATE_ATOM, $authHeader->Expires);

    // Our cache ID is our classname + our machine name.
    // @TODO: standardize this in Clients module.
    $cid = 'clients_connection_ms_dynamics_soap:' . $this->name;
    cache_set($cid, $authHeader, 'cache',  $expiry_date->getTimestamp());

    $this->isLoggedIn = TRUE;
    $status = TRUE;
    return $status;
  }

  /**
   * Get dates to pass to XRM.
   *
   * @return
   *  A flat array containing two date strings suitable for passing to XRM:
   *  - the current date.
   *  - the date for 1 day from now.
   */
  function getDates() {
    // XRM needs dates in UTC, otherwise it may think that we are giving it
    // dates in the past.
    $current_date = new DateTime('now', new DateTimeZone('UTC'));
    $current_date_string = $current_date->format("Y-m-d\TH:i:s") . ".00";

    $next_day_date = new DateTime('now +1 day', new DateTimeZone('UTC'));
    $next_day_date_string = $next_day_date->format("Y-m-d\TH:i:s"). ".00";

    return array($current_date_string, $next_day_date_string);
  }

  /**
   * Check the XML response to see if there is an error.
   *
   * @todo Expand this so that query classes can use it.
   *
   * @param $xml
   *  A SimpleXMLElement object for the response XML.
   * @param $namespace_uri
   *  The namespace UI to use for looking for the Fault XML elements. Because
   *  @£$%^ Microsoft send us XML back with this in different namespaces
   *  depending on when it's happening.
   *
   * @throws Exception
   *  Throws an exception if there is an error. The message of the exception is
   *  the fault text from the returned XML. It is advised to catch this and
   *  throw another exception with more detail about where the error occurred.
   */
  function checkXMLResponse($xml, $namespace_uri) {
    // See if there is an error.
    // MAY look like this:
    /*
    <Fault xmlns="http://schemas.microsoft.com/ws/2005/05/envelope/none">
      <Code>
        <Value>Receiver</Value>
        <Subcode>
          <Value xmlns:a="http://schemas.microsoft.com/net/2005/12/windowscommunicationfoundation/dispatcher">a:InternalServiceFault</Value>
        </Subcode>
      </Code>
      <Reason>
        <Text xml:lang="en-US">The server was unable to process the request due to an internal error.  For more information about the error, either turn on IncludeExceptionDetailInFaults (either from ServiceBehaviorAttribute or from the &lt;serviceDebug&gt; configuration behavior) on the server in order to send the exception information back to the client, or turn on tracing as per the Microsoft .NET Framework 3.0 SDK documentation and inspect the server trace logs.</Text>
      </Reason>
    </Fault>

    .. but might also be buried inside a SOAP envelope! THANKS MICROSOFT.
    */

    $xml->registerXPathNamespace('ns', $namespace_uri);
    $text_element = $xml->xpath("/ns:Fault/ns:Reason/ns:Text");
    if (!empty($text_element)) {
      $text = (string) $text_element[0];

      throw new Exception($text);
    }
  }

  /**
   * Generate a standard GUID.
   *
   * Cribbed from http://php.net/manual/en/function.com-create-guid.php
   */
  function generateGUID() {
    if (function_exists('com_create_guid') === TRUE) {
      return trim(com_create_guid(), '{}');
    }

    return sprintf('%04X%04X-%04X-%04X-%04X-%04X%04X%04X', mt_rand(0, 65535), mt_rand(0, 65535), mt_rand(0, 65535), mt_rand(16384, 20479), mt_rand(32768, 49151), mt_rand(0, 65535), mt_rand(0, 65535), mt_rand(0, 65535));
  }

  /**
   * Get the header XML for a query.
   *
   * @param $action
   *  The action to use for the query. One of:
   *    - 'RetrieveMultiple'
   *    - 'Retrieve'
   *
   * @return
   *  An XRML string for the query, starting with the opening 'Envelope' tag,
   *  and containing a complete 'Header' element. It is up to the caller to
   *  close the 'Envelope' element.
   */
  public function get_query_header($action) {
    // REFACTOR NEEDED

    $messageID = $this->generateGUID();

    list($currentTime, $nextDayTime) = $this->getDates();

    // The use of data in $this->credentials requires us to have called
    // connect() on the connection first!

    $headerXML = '
                    <s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
                    xmlns:a="http://www.w3.org/2005/08/addressing"
                    xmlns:u="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
                      <s:Header>
                        <a:Action s:mustUnderstand="1">
                        http://schemas.microsoft.com/xrm/2011/Contracts/Services/IOrganizationService/' . $action . '</a:Action>
                        <a:MessageID>
                        urn:uuid:' . $messageID . '</a:MessageID>
                        <a:ReplyTo>
                          <a:Address>
                          http://www.w3.org/2005/08/addressing/anonymous</a:Address>
                        </a:ReplyTo>
                        <VsDebuggerCausalityData xmlns="http://schemas.microsoft.com/vstudio/diagnostics/servicemodelsink">
                        uIDPozJEz+P/wJdOhoN2XNauvYcAAAAAK0Y6fOjvMEqbgs9ivCmFPaZlxcAnCJ1GiX+Rpi09nSYACQAA</VsDebuggerCausalityData>
                        <a:To s:mustUnderstand="1">' . $this->endpoint . '</a:To>
                        <o:Security s:mustUnderstand="1"
                        xmlns:o="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd">
                          <u:Timestamp u:Id="_0">
                            <u:Created>' . $currentTime . 'Z</u:Created>
                            <u:Expires>' . $nextDayTime . 'Z</u:Expires>
                          </u:Timestamp>
                          <EncryptedData Id="Assertion0"
                          Type="http://www.w3.org/2001/04/xmlenc#Element"
                          xmlns="http://www.w3.org/2001/04/xmlenc#">
                            <EncryptionMethod Algorithm="http://www.w3.org/2001/04/xmlenc#tripledes-cbc">
                            </EncryptionMethod>
                            <ds:KeyInfo xmlns:ds="http://www.w3.org/2000/09/xmldsig#">
                              <EncryptedKey>
                                <EncryptionMethod Algorithm="http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p">
                                </EncryptionMethod>
                                <ds:KeyInfo Id="keyinfo">
                                  <wsse:SecurityTokenReference xmlns:wsse="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd">

                                    <wsse:KeyIdentifier EncodingType="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
                                    ValueType="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509SubjectKeyIdentifier">
                                    ' . $this->credentials['identifier_key'] . '</wsse:KeyIdentifier>
                                  </wsse:SecurityTokenReference>
                                </ds:KeyInfo>
                                <CipherData>
                                  <CipherValue>' . $this->credentials['security_token0'] . '</CipherValue>
                                </CipherData>
                              </EncryptedKey>
                            </ds:KeyInfo>
                            <CipherData>
                              <CipherValue>' . $this->credentials['security_token1'] . '</CipherValue>
                            </CipherData>
                          </EncryptedData>
                        </o:Security>
                      </s:Header>
                      ';

    return $headerXML;
  }

  /**
   * Use curl to fetch POST SOAP request and fetch response
   * amended to force SSL version and set FOLLOWLOCATION
   */
  function GetSOAPResponse($postUrl, $request) {
    // Set up headers.
    $domain = parse_url($postUrl, PHP_URL_HOST);
    $headers = array(
      "Host: " . $domain,
      'Connection: Keep-Alive',
      "Content-type: application/soap+xml; charset=UTF-8",
    );

    $cURLHandle = curl_init();
    // Request that MS Dynamics send us gzipped data, which considerably speeds
    // up data transfer. Should the instance of MSD not support this, it falls
    // back to normal data (supposedly!).
    // Note that curl takes care of unzipping the data.
    curl_setopt($cURLHandle, CURLOPT_ENCODING, "gzip");
    curl_setopt($cURLHandle, CURLOPT_URL, $postUrl);
    curl_setopt($cURLHandle, CURLOPT_RETURNTRANSFER, 1);
    curl_setopt($cURLHandle, CURLOPT_TIMEOUT, variable_get('clients_connection_ms_dynamics_soap_curl_timeout', 120));
    curl_setopt($cURLHandle, CURLOPT_SSL_VERIFYPEER, FALSE);
    curl_setopt($cURLHandle, CURLOPT_FOLLOWLOCATION, TRUE);
    curl_setopt($cURLHandle, CURLOPT_SSLVERSION, 3);
    curl_setopt($cURLHandle, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    curl_setopt($cURLHandle, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($cURLHandle, CURLOPT_CUSTOMREQUEST, "POST");
    curl_setopt($cURLHandle, CURLOPT_POSTREDIR, 3);
    curl_setopt($cURLHandle, CURLOPT_POSTFIELDS, $request);
    $cookie_file = drupal_tempnam('temporary://', 'curl_cookie');
    curl_setopt($cURLHandle, CURLOPT_COOKIEJAR, $cookie_file);
    $response = curl_exec($cURLHandle);
    curl_close($cURLHandle);

    return $response;
  }

}
